//
//   This file is part of the KVIrc+ irc client distribution
//   Copyright (C) 1998 Szymon Stefanek (stefanek@tin.it)
//
//   This program is free software; you can redistribute it and/or
//   modify it under the terms of the GNU Library General Public
//   License as published by the Free Software Foundation; either
//   version 2 of the License, or (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//   Library General Public License for more details.
//
//   You should have received a copy of the GNU Library General Public License
//   along with this library; see the file COPYING.LIB.  If not, write to
//   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//   Boston, MA 02111-1307, USA.
//

#define _KVI_DEBUG_CLASS_NAME_ "KviUserParser"

#include "kvi_app.h"

#include <kfiledialog.h>
#include "kvi_ascii.h"
#include "kvi_dns.h"
#include "kvi_cache.h"
#include "kvi_token.h"

#define _macro_skipSenselessChars(_chptr) while((*_chptr) && ((*_chptr==' ')||(*_chptr=='\n')||(*_chptr=='\t')) )_chptr++
#define _macro_skipUpToEndOfCmd(_chptr) while((*_chptr) && (*_chptr!=';') && (*_chptr!=KVI_CHAR_END))_chptr++
#define _macro_getCommandParameters(__str,__lPtr,__rPtr,__pWnd,__pPar) QString __str(__lPtr,(__rPtr-__lPtr)+1); \
							m_lpIdentifiers->substitute(__pWnd,__str,__pPar);__str=__str.stripWhiteSpace()

#define _KVI_ERR_Success                0 /* Success                                 */
#define _KVI_ERR_NoPointer              1 /* NULL Pointer                            */
#define _KVI_ERR_EndAfterOpenBrace      2 /* End after open brace                    */
#define _KVI_ERR_MissingRightBrace      3 /* Missing right brace                     */
#define _KVI_ERR_ExpectedRightBrace     4 /* There is a different char               */
#define _KVI_ERR_MissingSeparator       5 /* Missing ;...there is something else     */
#define _KVI_ERR_BlockWithoutOpenBrace  6 /* Called parseBlock without an open brace */
#define _KVI_ERR_UnexpectedEndInCmd     7 /* parseCommand found an unexpected end    */
#define _KVI_ERR_UnknownCommand         8 /* unhandled command...pC points to it...  */
#define _KVI_ERR_MissingChannel         9 /* missing channel name(s)                 */
#define _KVI_ERR_NotConnected          10 /* Can't send data                         */
#define _KVI_ERR_MissingTarget         11 /* Missing target nickname                 */
#define _KVI_ERR_MissingData           12 /* Missing data to send...(RAW)            */
#define _KVI_ERR_MissingNick           13 /* Missing nickname                        */
#define _KVI_ERR_UnexpectedChar        14 /* Unexpected right brace                  */
#define _KVI_ERR_MissingVariable       15 /* Missing variable name (SET & UNSET)     */
#define _KVI_ERR_ExpectedVariable      16 /* Expected '%'...found something else     */
#define _KVI_ERR_MissingValue          17 /* Variable value expected                 */
#define _KVI_ERR_CantUnsetTwo          18 /* Can't unset two variables               */
#define _KVI_ERR_WindowNotFound        19 /* Window not found                        */
#define _KVI_ERR_MissingWindow         20 /* Missing window name                     */
#define _KVI_ERR_MissingText           21 /* Missing text to ECHO                    */
#define _KVI_ERR_NotOnChannel          22 /* You 're not on a channel                */
#define _KVI_ERR_MissingMode           23 /* Missing mode string                     */
#define _KVI_ERR_MissingCtcpType       24 /* Missing CTCP type                       */
#define _KVI_ERR_InfiniteRecursion     25 /* Infinite SET recursion                  */
#define _KVI_ERR_MissingExpression     26 /* Missing expression in if command        */
#define _KVI_ERR_UnexpectedEndInExpr   27 /* Unexpected end in expression            */
#define _KVI_ERR_MissingCommand        28 /* Missing command in if evaluation        */
#define _KVI_ERR_NotMatchingPrnths     29 /* Not mathing parenthesis                 */
#define _KVI_ERR_OperandDontMatchOprtr 30 /* Operand don't match operator            */
#define _KVI_ERR_UnknownOperator       31 /* Unknown operator                        */
#define _KVI_ERR_CantSetSystemVar      32 /* Cannot SET/UNSET a system variable      */
#define _KVI_ERR_ThisIsNotAnAlias      33 /* There are no %0,%1,%2 vars              */
#define _KVI_ERR_NonChannelAction      34 /* ME works only on a channel or query     */
#define _KVI_ERR_MissingTimerName      35 /* Timer name missing                      */
#define _KVI_ERR_UnknownTimerCmd       36 /* Unknown timer command                   */
#define _KVI_ERR_MissingTimerTime      37 /* Missing timer timeout                   */
#define _KVI_ERR_MissingTimerCommand   38 /* Missing timer command to execute        */
#define _KVI_ERR_TooSmallTimeout       39 /* Timeout less that 100 msec is DANGEROUS */
#define _KVI_ERR_MissingProcName       40 /* Missing process name                    */
#define _KVI_ERR_MissingProcPid        41 /* Missing process pid                     */
#define _KVI_ERR_Halt                  42 /* Not an error in fact , just HALT :)     */
#define _KVI_ERR_MissingServer         43 /* Missing server name                     */
#define _KVI_ERR_UnknownWindowCommand  44 /* Unknown window command                  */
#define _KVI_ERR_WrongWindowName       45 /* The window name does't start with a '@' */
#define _KVI_ERR_NoHostToResolve       46 /* No host to resolve                      */
#define _KVI_ERR_CantStartResolver     47 /* Can't start the kvidns program          */
#define _KVI_ERR_MissingFileName       48 /* Missing file name                       */
#define _KVI_ERR_MissingOperation      49 /* Missing operation                       */
#define _KVI_ERR_MissingKey            50 /* Missing key                             */
#define _KVI_ERR_UnknownCacheOperation 51 /* Unknown cache operation                 */

#define _KVI_ERR_ErrorMask            255 /* ErrorMask (numError&_KVI_ERR_ErrorMask) */

#define _KVI_CMD_parseTopLevel      (0<<8) /*:)))*/
#define _KVI_CMD_parseUserCommand   (1<<8) /*256 */
#define _KVI_CMD_parseBlock         (2<<8) /*512 */
#define _KVI_CMD_parseCommand       (3<<8) /*768 */
#define _KVI_CMD_parseJoin          (4<<8) /*1024*/
#define _KVI_CMD_parsePart          (5<<8)
#define _KVI_CMD_parseChat          (6<<8)
#define _KVI_CMD_parseQuery         (7<<8)
#define _KVI_CMD_parseRaw           (8<<8)
#define _KVI_CMD_parseNick          (9<<8)
#define _KVI_CMD_parseWhois        (10<<8)
#define _KVI_CMD_parseSet          (11<<8)
#define _KVI_CMD_parseUnset        (12<<8)
#define _KVI_CMD_parseEcho         (13<<8)
#define _KVI_CMD_parseTopic        (14<<8)
#define _KVI_CMD_parseMode         (15<<8)
#define _KVI_CMD_parseOp           (16<<8)
#define _KVI_CMD_executeMode       (17<<8)
#define _KVI_CMD_parseDeop         (18<<8)
#define _KVI_CMD_parseBan          (19<<8)
#define _KVI_CMD_parseUnban        (20<<8)
#define _KVI_CMD_parseVoice        (21<<8)
#define _KVI_CMD_parseDevoice      (22<<8)
#define _KVI_CMD_parseKick         (23<<8)
#define _KVI_CMD_parseSay          (24<<8)
#define _KVI_CMD_parseCtcp         (25<<8)
#define _KVI_CMD_executeAlias      (26<<8)
#define _KVI_CMD_parseIf           (27<<8)
#define _KVI_CMD_evaluateBoolean   (28<<8)
#define _KVI_CMD_parseHelp         (29<<8)
#define _KVI_CMD_parseSend         (30<<8)
#define _KVI_CMD_executeEvent      (31<<8)
#define _KVI_CMD_parseKban         (32<<8)
#define _KVI_CMD_parseWhowas       (33<<8)
#define _KVI_CMD_parseUserlist     (34<<8)
#define _KVI_CMD_parseClonescan    (35<<8)
#define _KVI_CMD_parseMe           (36<<8)
#define _KVI_CMD_parseTimer        (37<<8)
#define _KVI_CMD_parseRun          (38<<8)
#define _KVI_CMD_parseExec         (39<<8)
#define _KVI_CMD_parseKillproc     (40<<8)
#define _KVI_CMD_parseWriteproc    (41<<8)
#define _KVI_CMD_parseProclist     (42<<8)
#define _KVI_CMD_parseWhile        (43<<8)
#define _KVI_CMD_parsePlay         (44<<8)
#define _KVI_CMD_parseHalt         (45<<8)
#define _KVI_CMD_executePopup      (46<<8)
#define _KVI_CMD_parseServer       (47<<8)
#define _KVI_CMD_parseNotice       (48<<8)
#define _KVI_CMD_parseAway         (49<<8)
#define _KVI_CMD_parseBack         (50<<8)
#define _KVI_CMD_parseMsg          (51<<8)
#define _KVI_CMD_parseWindow       (52<<8)
#define _KVI_CMD_parseForeach      (53<<8)
#define _KVI_CMD_parseDns          (54<<8)
#define _KVI_CMD_parseAscii        (55<<8)
#define _KVI_CMD_parseCache        (56<<8)

#define _KVI_ERR_WhereMask          (~255)

//================ handlers Table ==================//
parseCmdTable KviUserParser::cmdTab[]={
	{ "OP"      , 2 , &KviUserParser::parseOp      },{ "DEOP"     , 4 , &KviUserParser::parseDeop      },
	{ "KICK"    , 4 , &KviUserParser::parseKick    },{ "SAY"      , 3 , &KviUserParser::parseSay       },
	{ "CHAT"    , 4 , &KviUserParser::parseChat    },{ "QUERY"    , 5 , &KviUserParser::parseQuery     },
	{ "RAW"     , 3 , &KviUserParser::parseRaw     },{ "NICK"     , 4 , &KviUserParser::parseNick      },
	{ "WHOIS"   , 5 , &KviUserParser::parseWhois   },{ "SET"      , 3 , &KviUserParser::parseSet       },
	{ "UNSET"   , 5 , &KviUserParser::parseUnset   },{ "ECHO"     , 4 , &KviUserParser::parseEcho      },
	{ "TOPIC"   , 5 , &KviUserParser::parseTopic   },{ "MODE"     , 4 , &KviUserParser::parseMode      },
	{ "JOIN"    , 4 , &KviUserParser::parseJoin    },{ "PART"     , 4 , &KviUserParser::parsePart      },
	{ "BAN"     , 3 , &KviUserParser::parseBan     },{ "UNBAN"    , 5 , &KviUserParser::parseUnban     },
	{ "VOICE"   , 5 , &KviUserParser::parseVoice   },{ "DEVOICE"  , 7 , &KviUserParser::parseDevoice   },
	{ "CTCP"    , 4 , &KviUserParser::parseCtcp    },{ "IF"       , 2 , &KviUserParser::parseIf        },
	{ "HELP"    , 4 , &KviUserParser::parseHelp    },{ "SEND"     , 4 , &KviUserParser::parseSend      },
	{ "KBAN"    , 4 , &KviUserParser::parseKban    },{ "WHOWAS"   , 6 , &KviUserParser::parseWhowas    },
	{ "USERLIST", 8 , &KviUserParser::parseUserlist},{ "CLONESCAN", 9 , &KviUserParser::parseClonescan },
	{ "ME"      , 2 , &KviUserParser::parseMe      },{ "TIMER"    , 5 , &KviUserParser::parseTimer     },
	{ "RUN"     , 3 , &KviUserParser::parseRun     },{ "EXEC"     , 4 , &KviUserParser::parseExec      },
	{ "KILLPROC", 8 , &KviUserParser::parseKillproc},{ "WRITEPROC", 9 , &KviUserParser::parseWriteproc },
	{ "PROCLIST", 8 , &KviUserParser::parseProclist},{ "WHILE"    , 5 , &KviUserParser::parseWhile     },
	{ "PLAY"    , 4 , &KviUserParser::parsePlay    },{ "HALT"     , 4 , &KviUserParser::parseHalt      },
	{ "SERVER"  , 6 , &KviUserParser::parseServer  },{ "NOTICE"   , 6 , &KviUserParser::parseNotice    },
	{ "AWAY"    , 4 , &KviUserParser::parseAway    },{ "BACK"     , 4 , &KviUserParser::parseBack      },
	{ "MSG"     , 3 , &KviUserParser::parseSay     },{ "WINDOW"   , 6 , &KviUserParser::parseWindow    },
	{ "FOREACH" , 7 , &KviUserParser::parseForeach },{ "DNS"      , 3 , &KviUserParser::parseDns       },
	{ "ASCII"   , 5 , &KviUserParser::parseAscii   },{ "CACHE"    , 5 , &KviUserParser::parseCache     },
	{ 0         , 0 , 0                            }
};

//================ error Strings ====================//
const char * KviUserParser::errTab[]={
	"Error 0 (success ?)",                               "Internal error : NULL pointer in data",
	"Unexpected end of command after open brace",        "Unbalanced open brace",
	"Unexpected character, right brace expected",        "Unexpected character, command separator expected",
	"Internal error : parseBlock call without a block",  "Unexpected end of command",
	"Unknown command",                                   "Missing channel name",
	"Can't send data, no active connection",             "Missing target nickname",
	"No data to send",                                   "Missing nickname",
	"Unexpected character",                              "Missing variable name",
	"Unexpected character, variable identifier expected","Missing variable value",
	"Too many operands for unary operator UNSET",        "Window not found",
	"Missing window name",                               "Missing text",
	"Missing channel name,and you're not on a channel",  "Missing mode parameter",
	"Missing CTCP type",                                 "Infinite recursion in variable evaluation",
	"Missing expression in if command",                  "Unexpected end in expression",
	"Missing command",                                   "Not matching parenthesis",
	"Operand don't match operator",                      "Unknown operator",
	"Can not SET/UNSET a system variable",               "This is not an alias (there are no %n variables)",
	"ME works only on channel or query windows",         "Missing timer name",
	"Unknown timer command",                             "Missing timer timeout",
	"Missing timer command",                             "Too small timeout",
	"Missing process name",                              "Missing process pid",
	"Execution halted",                                  "Missing server name",
	"Unknown window operation",                          "Wrong window name. [Must start with a '*']",
	"No host to resolve",                                "Unable to start kvidns resolver program",
	"Missing file name",                                 "Missing operation",
	"Missing key",                                       "Unknown cache operation",
	0
};
//=============== error Location Table ================//
const char * KviUserParser::fncTab[]={
	"COMMAND PARSER",  "COMMAND PARSER",     "COMMAND BLOCK",
	"COMMAND",         "JOIN",               "PART",
	"CHAT",            "QUERY",              "RAW",
	"NICK",            "WHOIS",              "SET",
	"UNSET",           "ECHO",               "TOPIC",
	"MODE",            "OP",                 "MODE EXECUTE",
	"DEOP",            "BAN",                "UNBAN",
	"VOICE",           "DEVOICE",            "KICK",
	"SAY",             "CTCP",               "ALIAS EXECUTE",
	"IF",              "EVAL BOOLEAN EXPR",  "HELP",
	"SEND",            "EVENT EXECUTE",      "KBAN",
	"WHOWAS",          "USERLIST",           "CLONESCAN",
	"ME",              "TIMER",              "RUN",
	"EXEC",            "KILLPROC",           "WRITEPROC",
	"PROCLIST",        "WHILE",              "PLAY",
	"HALT",            "POPUP EXECUTE",      "SERVER",
	"NOTICE",          "AWAY",               "BACK",
	"MSG",             "WINDOW",             "FOREACH",
	"DNS",             "ASCII",              "CACHE",
	0
};
//============ constructor & destructor ============//
KviUserParser::KviUserParser()
{
	m_lpVarList=new QList<KviVariable>;
	m_lpVarList->setAutoDelete(true);
	m_lpCache=new KviCache();
}
KviUserParser::~KviUserParser()
{
	while(!m_lpVarList->isEmpty())m_lpVarList->removeLast();
	delete m_lpVarList;
	delete m_lpIdentifiers;
	delete m_lpTimerManager;
	delete m_lpHelp;
	delete m_lpProcessManager;
	delete m_lpCache;
}
//============ activate ============//
void KviUserParser::activate(KviFrame *lpFrame)
{
	_debug_entertrace("activate");
	m_lpFrm         = lpFrame;
	m_lpSock        = lpFrame->m_lpSock;
	m_lpMdi         = lpFrame->m_lpMdi;
	m_lpCons        = lpFrame->m_lpConsole;
	m_lpGlb         = lpFrame->m_lpGlb;
	m_lpOpt         = lpFrame->m_lpOpt;
	m_lpAliasManager= lpFrame->m_lpAliasManager;
	m_lpProcessManager= new KviProcessManager(lpFrame);
	m_lpHelp        = new KviHelp();
	m_lpAliasList   = m_lpAliasManager->m_lpAliasList;
	m_lpIdentifiers = new KviIdentifiers(lpFrame,this);
	m_lpTimerManager= new KviTimerManager(this);
	m_szLastDCCTarget   ="";
	m_szLastQueryTarget ="";
	_debug_leavetrace("activate");
}
//============ handleError ============//
void KviUserParser::handleError(KviMdiChild *lpC,int numErr,const char *pC)
{
	_debug_entertrace("handleError");
	int location=((numErr & _KVI_ERR_WhereMask)>>8);
	int errorNum=( numErr & _KVI_ERR_ErrorMask);
	if(errorNum==_KVI_ERR_Halt)return;
	if(pC && (*pC)){
		QString lastChars(pC,10);
		lastChars+="...";
		lpC->doFmtOutput(KVI_OUT_ERROR,"[%s]:[%s]: %s",fncTab[location],lastChars.data(),errTab[errorNum]);
	} else lpC->doFmtOutput(KVI_OUT_ERROR,"[%s]: %s",fncTab[location],errTab[errorNum]);
	_debug_leavetrace("handleError");
}
//============ parseUserInput ============//
void KviUserParser::parseUserInput(KviMdiChild *lpC,int winType,const char *winName,QString &szData)
{
	_debug_entertrace("parseUserInput");
	if(szData.isEmpty())return;
	const char *pC=szData.data();
	_macro_skipSenselessChars(pC);
	if(! (*pC)) return;
	if( ((*pC)=='/') || ((*pC)==m_lpOpt->chPersonalCommandPrefix) ){
		pC++;
		if(!(*pC))return;
		int numErr=parseUserCommand(lpC,&pC,0);
		if((*pC==KVI_CHAR_END)&&(!(numErr & _KVI_ERR_ErrorMask)))numErr=(_KVI_CMD_parseTopLevel|_KVI_ERR_UnexpectedChar);
		if(numErr & _KVI_ERR_ErrorMask)handleError(lpC,numErr,pC);
		return;
	} else {
		if(!((m_lpSock->m_bConnected)||(winType == KVI_WND_TYPE_USER)||(winType==KVI_WND_TYPE_CHAT))){
			lpC->doOutput(KVI_OUT_ERROR,"You're not connected to a server");
			return;
		}
	}
	m_lpIdentifiers->substitute(lpC,szData,0);
	switch(winType){
	case KVI_WND_TYPE_CHAN:
	case KVI_WND_TYPE_QUERY:
		{
			QString szNicko;
			m_lpFrm->formatNickOutputString(szNicko,m_lpGlb->szNick);
			lpC->doFmtOutput((winType==KVI_WND_TYPE_CHAN ? KVI_OUT_OWN : KVI_OUT_PRIVMSG),"%s%s",szNicko.data(),szData.data());
			m_lpSock->sendFmtData("PRIVMSG %s :%s",winName,szData.data());	
		}
		break;
	case KVI_WND_TYPE_CHAT:
		{
			if(!((KviChatWnd *)lpC)->m_bConnected)return;
			lpC->sendData(szData.data());
			QString szNicko;
			m_lpFrm->formatNickOutputString(szNicko,m_lpGlb->szNick);	
			lpC->doFmtOutput(KVI_OUT_OWN,"%s%s",szNicko.data(),szData.data());
		}
		break;		
	case KVI_WND_TYPE_STATUS:
		lpC->doFmtOutput(KVI_OUT_RAW,"[RAW TO SERVER] : %s",szData.data());
		m_lpSock->sendData(szData.data());	
		break;
	case KVI_WND_TYPE_USER:
		if(m_lpFrm->m_lpEventManager->lpEvent[KVI_Event_OnUserWndInput]->bEnabled){
			QString szPara=lpC->name();
			szPara+=' ';
			szPara+=szData;
			executeEvent(lpC,m_lpFrm->m_lpEventManager->lpEvent[KVI_Event_OnUserWndInput],szPara);
		}
		break;
	}
	_debug_leavetrace("parseUserInput");
}
//============ parseUserCommand ============//
int KviUserParser::parseUserCommand(KviMdiChild *lpC,const char ** pC,KviAliasParams *pP)
{
	_debug_entertrace("parseUserCommand");
	//  **pC is the character
	//   *pC is the pointer to character
	//    pC is the pointer to the pointer to character
	if(!(*pC))return (_KVI_CMD_parseUserCommand|_KVI_ERR_NoPointer);	//empty string
	_macro_skipSenselessChars((*pC));
	if(! (**pC) )return (_KVI_CMD_parseUserCommand|_KVI_ERR_Success);
	if( (**pC)==KVI_CHAR_BEGIN ){
		int numError=parseBlock(lpC,pC,pP);
		if(numError & _KVI_ERR_ErrorMask)return numError;
	} else {
		int numError=parseCommand(lpC,pC,pP);
		if(numError & _KVI_ERR_ErrorMask)return numError;
	}
	_macro_skipSenselessChars((*pC));
	if(!(**pC))return (_KVI_CMD_parseUserCommand|_KVI_ERR_Success);
	if((**pC) != KVI_CHAR_END){
		if((**pC) != ';')return (_KVI_CMD_parseUserCommand|_KVI_ERR_MissingSeparator);
		(*pC)++;
		_macro_skipSenselessChars((*pC));
		if((**pC) == KVI_CHAR_END){
			return (_KVI_CMD_parseUserCommand|_KVI_ERR_Success);
		}
	} else {
		return (_KVI_CMD_parseUserCommand|_KVI_ERR_Success);
	}
	return parseUserCommand(lpC,pC,pP);
	_debug_leavetrace("parseUserCommand");
}
//============ parseBlock ============//
int KviUserParser::parseBlock(KviMdiChild *lpC,const char ** pC,KviAliasParams *pP)
{
	_debug_entertrace("parseBlock");
	if((**pC) != KVI_CHAR_BEGIN)return (_KVI_CMD_parseBlock|_KVI_ERR_BlockWithoutOpenBrace);
	(*pC)++;
	if(!(**pC))return (_KVI_CMD_parseBlock|_KVI_ERR_EndAfterOpenBrace);
	int numError=parseUserCommand(lpC,pC,pP);
	if(numError & _KVI_ERR_ErrorMask)return numError;
	_macro_skipSenselessChars((*pC));
	if(!(**pC))return (_KVI_CMD_parseBlock|_KVI_ERR_MissingRightBrace); //here
	if(**pC != KVI_CHAR_END)return (_KVI_CMD_parseBlock|_KVI_ERR_ExpectedRightBrace);
	(*pC)++;
	return (_KVI_CMD_parseBlock|_KVI_ERR_Success);
	_debug_leavetrace("parseBlock");
}
//============ parseCommand ============//
int KviUserParser::parseCommand(KviMdiChild *lpC,const char ** pC,KviAliasParams *pP)
{
	_debug_entertrace("parseCommand");
	if(!(*pC))return (_KVI_CMD_parseCommand|_KVI_ERR_NoPointer);
	if(!(**pC))return (_KVI_CMD_parseCommand|_KVI_ERR_UnexpectedEndInCmd);
	//Find the built in command...look un the handler proc in table
	if(**pC=='$'){ //First substitute the first word (a variable command)
		const char *pidx=(*pC);
		while((*pidx)&&(*pidx!=' ')&&(*pidx!='\t')&&(*pidx!='\n')&&(*pidx!=';'))pidx++;
		QString szC;
		if(!pidx)szC=QString(*pC);
		else szC=QString((*pC),(pidx-(*pC))+1);
		m_lpIdentifiers->substitute(lpC,szC,pP);
		szC=szC.stripWhiteSpace();
		const char *ptr=szC.data();
		for (int i=0;cmdTab[i].szName!=0;i++){
			if (!strcasecmp((ptr),cmdTab[i].szName)){
				if(pidx)(*pC)=pidx;
				else (*pC)+=strlen((*pC));
				_macro_skipSenselessChars((*pC));
				return (this->*(cmdTab[i].proc))(lpC,pC,pP);
			}
		}
		//Not found...look for an alias
		KviAliasStruct *lpA;
		for(lpA=m_lpAliasList->first();lpA;lpA=m_lpAliasList->next()){
			if(!strcasecmp((ptr),lpA->szName.data())){
				if(pidx)(*pC)=pidx;
				else (*pC)+=strlen((*pC));
				_macro_skipSenselessChars((*pC));
				QString szParams="";
				if(**pC){
					const char *rghPtr=(*pC);
					_macro_skipUpToEndOfCmd(rghPtr);
					szParams=QString((*pC),(rghPtr-(*pC))+1);
					m_lpIdentifiers->substitute(lpC,szParams,pP);
					(*pC)=rghPtr;
				}
				return executeAlias(lpC,lpA,szParams);
			}
		}
		//Not found...Break the executing chain...
		return (_KVI_CMD_parseCommand|_KVI_ERR_UnknownCommand);
	}
	//Normal command
	int nLen=strlen(*pC);
	const char *pSpace=0;
	for (int i=0;cmdTab[i].szName!=0;i++){
		if(nLen>=cmdTab[i].lenName){
			pSpace=(*pC)+cmdTab[i].lenName;
			if((*pSpace==' ')||(*pSpace=='\t')||(*pSpace=='\n')||(!(*pSpace))||(*pSpace==';')){
				if (!strncasecmp((*pC),cmdTab[i].szName,cmdTab[i].lenName)){
					(*pC)=pSpace;
					_macro_skipSenselessChars((*pC));
		 			return (this->*(cmdTab[i].proc))(lpC,pC,pP);
				}
			}
		}
	}
	//Not found...look for an alias
	KviAliasStruct *lpA;
	for(lpA=m_lpAliasList->first();lpA;lpA=m_lpAliasList->next()){
		int iLen=strlen(lpA->szName.data());
		if(!strncasecmp((*pC),lpA->szName.data(),iLen)){
			const char *ptrToEnd=(*pC)+iLen;
			if((*ptrToEnd==' ')||(*ptrToEnd=='\n')||(*ptrToEnd=='\t')||(*ptrToEnd==';')||(!(*ptrToEnd))){
				(*pC)=ptrToEnd;
				_macro_skipSenselessChars((*pC));
				QString szParams="";
				if(**pC){
					const char *rghPtr=(*pC);
					_macro_skipUpToEndOfCmd(rghPtr);
					szParams=QString((*pC),(rghPtr-(*pC))+1);
					m_lpIdentifiers->substitute(lpC,szParams,pP);
					*pC=rghPtr;
				}
				return executeAlias(lpC,lpA,szParams);
			}
		}
	}
	//Not found...Break the executing chain...
	return (_KVI_CMD_parseCommand|_KVI_ERR_UnknownCommand);
	_debug_leavetrace("parseCommand");
}
//============ executeAlias ============//
int KviUserParser::executeAlias(KviMdiChild *lpC,KviAliasStruct *lpA,QString &szParams)
{
	_debug_entertrace("executeAlias");
	KviAliasParams szAliasParams;
	if(szParams.isEmpty()){
		for(int i=0;i<10;i++)szAliasParams.param[i]=KVI_STR_NULL;
	} else {
		szAliasParams.param[0]=szParams.stripWhiteSpace();
		for(int i=1;i<10;i++){
			szParams=szParams.stripWhiteSpace();	
			if(!szParams.isEmpty()){
				int idx=szParams.find(' ');
				if((idx==-1)||(i==9)){szAliasParams.param[i]=szParams.data();szParams="";}
				else {szAliasParams.param[i]=szParams.left(idx);szParams.remove(0,idx+1);}
			} else szAliasParams.param[i]=KVI_STR_NULL;
		}
	}
	register const char *pCh=lpA->szBuffer.data();
	const char *pBegin=0;
	QString *pAliasBuffer=new QString(1);
	while(*pCh){
		while((*pCh)&&((*pCh==' ')||(*pCh=='\n')||(*pCh=='\t')))pCh++;
		if(*pCh){
			if(*pCh=='@')while((*pCh)&&(*pCh != '\n'))pCh++;
			else { 
				while(*pCh && (*pCh != '@')){
					pBegin=pCh;
					while((*pCh) && (*pCh!='\n'))pCh++;
					if(*pCh){
						pAliasBuffer->append(QString(pBegin,(pCh-pBegin)+1).data());
						pBegin=pCh;
						while((*pCh)&&((*pCh==' ')||(*pCh=='\n')||(*pCh=='\t')))pCh++;
						if(pBegin != pCh)pAliasBuffer->append(QString(pBegin,(pCh-pBegin)+1).data());		
					} else { //if it was the end copy the remaining
						if(pCh != pBegin)pAliasBuffer->append(QString(pBegin,(pCh-pBegin)+1).data());
					}
				}
			}
		}
	}
	if(!pAliasBuffer->isEmpty()){
		pCh=pAliasBuffer->data();
		_macro_skipSenselessChars(pCh);
		if((*pCh) != '/'){
			delete pAliasBuffer;
			return (_KVI_CMD_executeAlias|_KVI_ERR_UnexpectedChar);
		}
		pCh++;
		if(*pCh){
			int nErr= parseUserCommand(lpC,&pCh,&szAliasParams);
			delete pAliasBuffer;
			return nErr;
		} else {
			delete pAliasBuffer;
			return (_KVI_CMD_executeAlias|_KVI_ERR_UnexpectedEndInCmd);
		}
	}
	delete pAliasBuffer;
	_debug_leavetrace("executeAlias");
	return (_KVI_CMD_executeAlias|_KVI_ERR_Success);
}
//============ executeEvent ============//
int KviUserParser::executeEvent(KviMdiChild *lpC,KviEventStruct *lpE,QString &szParams)
{
	_debug_entertrace("executeEvent");
	KviAliasParams szAliasParams;
	if(szParams.isEmpty()){
		for(int i=0;i<10;i++)szAliasParams.param[i]=KVI_STR_NULL;
	} else {
		szAliasParams.param[0]=szParams.stripWhiteSpace();
		for(int i=1;i<10;i++){
			szParams=szParams.stripWhiteSpace();	
			if(!szParams.isEmpty()){
				int idx=szParams.find(' ');
				if((idx==-1)||(i==9)){szAliasParams.param[i]=szParams.data();szParams="";}
				else {szAliasParams.param[i]=szParams.left(idx);szParams.remove(0,idx+1);}
			} else szAliasParams.param[i]=KVI_STR_NULL;
		}
	}
	register const char *pC=lpE->szBuffer.data();
	const char *pBegin=0;
	QString szAliasBuffer="";
	while(*pC){
		while((*pC)&&((*pC==' ')||(*pC=='\n')||(*pC=='\t')))pC++;
		if(*pC){
			if(*pC=='@')while((*pC)&&(*pC != '\n'))pC++;
			else { 
				while(*pC && (*pC != '@')){
					pBegin=pC;
					while((*pC) && (*pC!='\n'))pC++;
					if(*pC){
						szAliasBuffer.append(QString(pBegin,(pC-pBegin)+1).data());
						pBegin=pC;
						while((*pC)&&((*pC==' ')||(*pC=='\n')||(*pC=='\t')))pC++;
						if(pBegin != pC)szAliasBuffer.append(QString(pBegin,(pC-pBegin)+1).data());		
					} else { //if it was the end copy the remaining
						if(pC != pBegin)szAliasBuffer.append(QString(pBegin,(pC-pBegin)+1).data());
					}
				}
			}
		}
	}
	if(!szAliasBuffer.isEmpty()){
		pC=szAliasBuffer.data();
		_macro_skipSenselessChars(pC);
		if((*pC) != '/'){
			m_lpCons->doFmtOutput(KVI_OUT_ERROR,"Error in event : %s",lpE->szName.data());	
			handleError(m_lpCons,_KVI_CMD_executeEvent|_KVI_ERR_UnexpectedChar,pC);
			return KVI_EventErr_Error;
		}
		pC++;
		if(*pC){
			if(*pC){
				int numErr=parseUserCommand(lpC,&pC,&szAliasParams);
				if(numErr & _KVI_ERR_ErrorMask){
					if((numErr & _KVI_ERR_ErrorMask)==_KVI_ERR_Halt){
						return KVI_EventErr_Halt; //Stop
					} else {
						m_lpCons->doFmtOutput(KVI_OUT_ERROR,"Error in event : %s",lpE->szName.data());		
						handleError(m_lpCons,numErr,pC);
						return KVI_EventErr_Error;
					}
				} else return KVI_EventErr_Success;
			}
		}
		m_lpCons->doFmtOutput(KVI_OUT_ERROR,"Error in event : %s",lpE->szName.data());
		handleError(m_lpCons,(_KVI_CMD_executeEvent|_KVI_ERR_UnexpectedEndInCmd),pC);
		return KVI_EventErr_Error;
	}
	_debug_leavetrace("executeEvent");
	return KVI_EventErr_Success;
}
//============ executePopup ============//
int KviUserParser::executePopup(KviMdiChild *lpC,QString &szAction)
{
	_debug_entertrace("executePopup");
	register const char *pC=szAction.data();
	const char *pBegin=0;
	QString szPopupBuffer="";
	while(*pC){
		while((*pC)&&((*pC==' ')||(*pC=='\n')||(*pC=='\t')))pC++;
		if(*pC){
			if(*pC=='@')while((*pC)&&(*pC != '\n'))pC++;
			else { 
				while(*pC && (*pC != '@')){
					pBegin=pC;
					while((*pC) && (*pC!='\n'))pC++;
					if(*pC){
						szPopupBuffer.append(QString(pBegin,(pC-pBegin)+1).data());
						pBegin=pC;
						while((*pC)&&((*pC==' ')||(*pC=='\n')||(*pC=='\t')))pC++;
						if(pBegin != pC)szPopupBuffer.append(QString(pBegin,(pC-pBegin)+1).data());		
					} else { //if it was the end copy the remaining
						if(pC != pBegin)szPopupBuffer.append(QString(pBegin,(pC-pBegin)+1).data());
					}
				}
			}
		}
	}
	if(!szPopupBuffer.isEmpty()){
		pC=szPopupBuffer.data();
		_macro_skipSenselessChars(pC);
		if((*pC) != '/'){
			handleError(lpC,_KVI_CMD_executePopup|_KVI_ERR_UnexpectedChar,pC);
			return KVI_EventErr_Error;
		}
		pC++;
		if(*pC){
			if(*pC){
				int numErr=parseUserCommand(lpC,&pC,0);
				if(numErr & _KVI_ERR_ErrorMask){
					if((numErr & _KVI_ERR_ErrorMask)==_KVI_ERR_Halt){
						return KVI_EventErr_Halt; //Stop
					} else {
						handleError(lpC,numErr,pC);
						return KVI_EventErr_Error;
					}
				} else return KVI_EventErr_Success;
			}
		}
		handleError(lpC,(_KVI_CMD_executePopup|_KVI_ERR_UnexpectedEndInCmd),pC);
		return KVI_EventErr_Error;
	}
	_debug_leavetrace("executePopup");
	return KVI_EventErr_Success;
}
//============ executeMode ===========//
int KviUserParser::executeMode(QString &szChan,QString &szTrg,bool bPlus,char chMode)
{
	_debug_entertrace("executeMode");
	QString szTar="";
	int num=0;
	QString szMod=( bPlus ? "+" : "-" );
	while(!szTrg.isEmpty()){
		int idx=szTrg.find(',');
		if(!szTar.isEmpty())szTar+=' ';
		if(idx==-1){ szTar+=szTrg; szTrg="";}
		else { szTar+=szTrg.left(idx); szTrg.remove(0,idx+1); }
		num++;
		szMod+=chMode;
		if(num==3){
			num=0;
			if(m_lpSock->sendFmtData("MODE %s %s %s",szChan.data(),szMod.data(),szTar.data()))return (_KVI_CMD_executeMode|_KVI_ERR_NotConnected);	
			szMod=( bPlus ? "+" : "-" ); szTar="";
		}
	}
	if(num)if(m_lpSock->sendFmtData("MODE %s %s %s",szChan.data(),szMod.data(),szTar.data()))return (_KVI_CMD_executeMode|_KVI_ERR_NotConnected);	
	return (_KVI_CMD_executeMode|_KVI_ERR_Success);
	_debug_leavetrace("executeMode");
}
//============ evaluateBoolean ============//

//<boolean>  ::= [!]<boolean>
//<boolean>  ::= (<boolean>)
//<boolean>  ::= <boolean> <op> <boolean>

int KviUserParser::evaluateBoolean(KviMdiChild *lpC,const char ** pC,bool *bResult,KviAliasParams *pP,bool hadBrace)
{
	_debug_entertrace("evaluateBoolean");
	_macro_skipSenselessChars((*pC));
	if(!(**pC))return (_KVI_CMD_evaluateBoolean|_KVI_ERR_UnexpectedEndInExpr);
	//First check for a NOT
	if(**pC=='!'){
		(*pC)++;
		_macro_skipSenselessChars((*pC));
		if(!(**pC))return (_KVI_CMD_evaluateBoolean|_KVI_ERR_UnexpectedEndInExpr);	
		int numErr=evaluateBoolean(lpC,pC,bResult,pP,hadBrace);
		if(numErr & _KVI_ERR_ErrorMask)return numErr;
		(*bResult) = !(*bResult);
	} else if(**pC=='('){
		(*pC)++;
		_macro_skipSenselessChars((*pC));
		if(!(**pC))return (_KVI_CMD_evaluateBoolean|_KVI_ERR_UnexpectedEndInExpr);	
		int numErr=evaluateBoolean(lpC,pC,bResult,pP,true);
		if(numErr & _KVI_ERR_ErrorMask)return numErr;
		if((**pC) != ')')return (_KVI_CMD_evaluateBoolean|_KVI_ERR_NotMatchingPrnths);
		(*pC)++;
		_macro_skipSenselessChars((*pC));
	} else { //first operand not in parenthesis
		//skip up to end,'&','|',')'
		const char *rghPtr=*pC;
		while((*rghPtr)&&(*rghPtr != '&')&&(*rghPtr != '|')&&(*rghPtr != ')'))rghPtr++;
		QString szOp((*pC),(rghPtr-(*pC))+1);
		szOp=szOp.stripWhiteSpace();
		if(szOp.isEmpty())return (_KVI_CMD_evaluateBoolean|_KVI_ERR_MissingExpression);
		m_lpIdentifiers->substitute(lpC,szOp,pP);
		(*pC)=rghPtr;
		rghPtr=szOp.data();
		(*bResult)= (	strcmp(rghPtr,KVI_STR_NULL)  &&
						strcmp(rghPtr,KVI_STR_FALSE) &&
						strcmp(rghPtr,"0") );
	}
	if(! (**pC))return (_KVI_CMD_evaluateBoolean|_KVI_ERR_Success);
	if(**pC==')'){
		if(hadBrace)return (_KVI_CMD_evaluateBoolean|_KVI_ERR_Success);
		else return (_KVI_CMD_evaluateBoolean|_KVI_ERR_NotMatchingPrnths);
	} else if((**pC=='&') || (**pC=='|')){
		char chOperator=**pC;
		(*pC)++;
		if(!(**pC))return (_KVI_CMD_evaluateBoolean|_KVI_ERR_UnexpectedEndInExpr);
		bool bResult2=false;
		int numErr=evaluateBoolean(lpC,pC,&bResult2,pP,hadBrace);
		if(numErr & _KVI_ERR_ErrorMask)return numErr;
		if(chOperator=='&')(*bResult)= ((*bResult) && bResult2);
		else (*bResult)= ((*bResult) || bResult2);
	} else return (_KVI_CMD_evaluateBoolean|_KVI_ERR_UnknownOperator);
	_debug_leavetrace("evaluateBoolean");
	return (_KVI_CMD_evaluateBoolean|_KVI_ERR_Success);
}
#include "kvi_uparser2.cpp"
